#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl my-check-gp.pl filename.gp...
#
# Check for some mistakes in a GP script file.

use 5.010;
use warnings;
use strict;
use File::Slurp;
use FindBin;
use Getopt::Long;
use List::Util 'max','min';

# uncomment this to run the ### lines
# use Smart::Comments;


our $VERSION = 0;

my $progname = $FindBin::Script;
my $verbose = 0;
{
  my $help = sub {
    print "$progname [--options] filename...\n";
    my @opts =
      (['-h, --help',    'Print this help'],
       ['-v, --version', 'Print program version'],
       ['--verbose',     'Print extra messages'],
      );
    my $width = 2 + max (map { length ($_->[0]) } @opts);
    foreach (@opts) {
      printf "%-*s%s\n", $width, $_->[0], $_->[1];
    }
    print "\n";
    exit 0;
  };

  GetOptions ('help|?' => $help,
              version => sub {
                print "$progname version $VERSION\n";
                exit 0;
              },
              verbose    => \$verbose,
             )
    or exit 1;

  @ARGV or $help->();
}

my $gp_word_re = qr/[A-Za-z_][A-Za-z_0-9]*/;
my $gp_comment_re = qr{/\*.*?\*/|\\\\[^\r\n]*};
my $gp_string_re = qr{"([^\\"]|\\")*"?};

'"xx"'=~ $gp_string_re or die;
"\"x\\\"x\""=~ $gp_string_re or die;

my $gp_filename;
my $bad = 0;
sub bad {
  my $linenum = shift;
  $bad = 1;
  print "$gp_filename:$linenum: ",@_,"\n";
}
sub bad_but_not_fatal {
  my $linenum = shift;
  print "$gp_filename:$linenum: ",@_,"\n";
}

while (@ARGV) {
  $gp_filename = shift @ARGV;
  check_file();
}

sub check_file {
  my $gp_str = File::Slurp::slurp($gp_filename);

  my $gp_str_sans_comments = $gp_str;
  $gp_str_sans_comments =~ s{$gp_comment_re}
                            {str_leave_newlines($&)}eg;

  my $gp_str_sans_strings = $gp_str;
  $gp_str_sans_strings =~ s{$gp_comment_re|($gp_string_re)}
                           {my $quote = ($1 ? '"' : '');
                            $quote.str_leave_newlines($&).$quote}eg;
  # print $gp_str_sans_strings;

  if ($verbose) {
    print "$gp_filename ",count_lines($gp_str)," lines\n";
  }


  #----------------------------------------------------------------------------
  # check for function redefinition

  my $count_definitions = 0;
  my %seen_function;
  {
    my %redefinition_ok = (addhelp => 1,
                           if      => 1,
                           local   => 1,
                           my      => 1,
                          );

    my $str = $gp_str_sans_strings;
    while ($str =~ /^($gp_word_re)\s*\(/mg) {
      my $name = $1;
      next if $redefinition_ok{$name};
      $count_definitions++;

      my $linenum = pos_linenum($str);
      if ($seen_function{$name}) {
        bad($linenum, "function redefinition $name, previous at line $seen_function{$name}");
      } else {
        $seen_function{$name} = $linenum;
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for addhelp() duplication or function not present

  my $count_addhelp = 0;
  {
    my $str = $gp_str_sans_strings;
    my %seen;
    while ($str =~ /addhelp\s*\(\s*($gp_word_re)/sg) {
      my $name = $1;
      $count_addhelp++;

      my $linenum = pos_linenum($str);
      if ($seen{$name}) {
        bad($linenum, "addhelp($name) duplicate, previous at line $seen{$name}");
      } else {
        $seen{$name} = $linenum;
      }

      if (! $seen{$name}) {
        bad($linenum, "addhelp($name) no such function");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for addhelp() without function name in first line

  {
    my $str = $gp_str_sans_comments;
    my %seen;
    while ($str =~ /addhelp\s*\(\s*($gp_word_re)\s*,\s*"([^\r\n]*)/sg) {
      my $name = $1;
      my $line = $2;
      if ($line !~ /\b\Q$name\E\b/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "addhelp($name) without $name in first line of help string");
      }
    }
  }

  #----------------------------------------------------------------------------
  if ($verbose) {
    print "  $count_definitions functions, $count_addhelp helps\n";
  }
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

# return $str with all characters except \n newlines deleted
sub str_leave_newlines {
  my ($str) = @_;
  $str =~ tr/\n//cd;
  return $str;
}


#------------------------------------------------------------------------------

exit ($bad ? 1 : 0);
